/*
 * modules.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <dlfcn.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <stdarg.h>

#include "icebuteo.h"
#include "modules.h"
#include "vars.h"
#include "parser.h"

#define DEBUG 0
#include "debug.h"

/* callback names */
const char *cb_name[CB_MAX + 1] = {
	"mod_initialize", "mod_cleanup",
	"mod_timeout", "mod_filedescriptor",
	"mod_get_timeout_ptr", "mod_get_fd_list_ptr",
	NULL
};


/*****************************************************************************
 * module_lookup
 */
static void *module_lookup(MODULE *mod, int mod_id, const char *symbol) {
	char *error = NULL;
	void *addr = NULL;

	assert(mod != NULL);
	assert(mod_id >= 0);
	assert(mod_id < mod->info_count);
	assert(symbol != NULL);

	if (mod->info[mod_id].name) {
		DBG_MSG(2, "Looking up \"%s\" in module \"%s\" ...",
				symbol, mod->info[mod_id].name);

		dlerror();
		addr = dlsym(mod->info[mod_id].dlh, symbol);
		if ((error = dlerror())) {
			fprintf(stderr, "Can't find \"%s\" in \"%s\": %s\n",
					symbol, mod->info[mod_id].name, error);
			return NULL;
		}
		DBG_MSG(2, "Looking up \"%s\" in module \"%s\" ... found at %p",
				symbol, mod->info[mod_id].name, addr);
	}
	return addr;
}


/*****************************************************************************
 * module_init
 */
MODULE * module_init(char *path) {
	int i;
	MODULE *mod = NULL;
	DBG_MSG(1, "Initializing ...");

	/* alloc memory */
	mod = (MODULE *) malloc(sizeof(MODULE));
	mod->info_count = cfg.max_modules;
	mod->path = strdup(path);
	mod->info = (struct s_mod_info *) malloc(sizeof(struct s_mod_info) * mod->info_count);

	/* mark all slots as empty */
	for (i = 0; i < mod->info_count; i++) {
		mod->info[i].name = NULL;
	}
	
	/* store config */
	mod->path = strdup(path);

	return mod;
}


/*****************************************************************************
 * module_destroy
 */
void module_destroy(MODULE *mod) {
	assert(mod != NULL);

	DBG_MSG(1, "Cleaning up ...");
	module_remove_all(mod);

	/* free memory */
	if (mod->info)
		free(mod->info);
	if (mod->path)
		free(mod->path);
	free(mod);
}


/*****************************************************************************
 * module_load_config
 */
int module_load_config(MODULE *mod) {
	char *fn = cfg.module_file;
	int r;
	FILE *f;
	P_FORMAT format[] = {
		P_STRING, P_INTEGER, P_STRING, 0
	};
	P_LINE *ln;

	assert(mod != NULL);
	DBG_MSG(1, "Loading modules configuration from \"%s\"", fn);
	if ((f = fopen(fn, "r"))) {
		while ((ln = parse_line(fn, f, format, -1))) {
			r = module_configure(mod, ln->token[0].val, ln->token[1].val_int, ln->token[2].val);
			parse_free_line(ln);
			if (r == -1)
				return -1;
		}
		if (fclose(f) != 0)
			perror(fn);
	} else {
		perror(fn);
		exit(1);
	}
	DBG_MSG(1, "Loading modules configuration from \"%s\" completed.", fn);
	return 0;
}

/*****************************************************************************
 * module_load
 */
int module_configure(MODULE *mod, char *mod_name, int var_offset, char *args) {
	int id = 0;
	int i;

	DBG_MSG(1, "Configuring module \"%s\" (args = \"%s\")", mod_name, args);
	assert(mod != NULL);
	assert(mod_name != NULL);
	assert(args != NULL);

	/* find empty slot */
	while (mod->info[id].name != NULL) {
		id++;
		if (id >= mod->info_count) {
			fprintf(stderr, "Can't load module \"%s\": Too many modules loaded.\n", mod_name);
			return -1;
		}
	}

	/* init module info */
	mod->info[id].name = strdup(mod_name);
	mod->info[id].var_offset = var_offset;
	mod->info[id].real_name = NULL;
	mod->info[id].args = strdup(args);
	mod->info[id].data = NULL;
	mod->info[id].cfg_fd = NULL;
	mod->info[id].cfg_timeout = NULL;
	mod->info[id].dlh = NULL;
	mod->info[id].cfg_fd_old = NULL;
	for (i = 0; i < CB_MAX; i++) {
		mod->info[id].callback[i] = NULL;
	}

	return id;
}

/*****************************************************************************
 * module_load
 */
int module_load(MODULE *mod, int mod_id) {
	char *fn, *error;

	assert(mod != NULL);
	assert(mod_id < mod->info_count);

	/* check if module is configured */
	if (!mod->info[mod_id].name) {
		return -1;
	}

	DBG_MSG(1, "Loading module \"%s\"(%d)", mod->info[mod_id].name, mod_id);

	/* build module filename */
	fn = (char*) malloc(strlen(mod->path) + strlen(mod->info[mod_id].name) + 10);
	sprintf(fn, "%s/%s.so", mod->path, mod->info[mod_id].name);

	/* load module */
	DBG_MSG(3, "real filename: \"%s\"", fn);
	if (!(mod->info[mod_id].dlh = dlopen(fn, RTLD_LAZY))) {
		fprintf (stderr, "Can't load module \"%s\": %s\n", mod->info[mod_id].name, dlerror());
		free(fn);
		module_remove(mod, mod_id);
		return -1;
	}
	free(fn);

	/* get module name */
	if (!(mod->info[mod_id].real_name = *(char**) module_lookup(mod, mod_id, "mod_name"))) {
		fprintf(stderr, "Can't read module's real name: %s\n", error);
		module_remove(mod, mod_id);
		return -1;
	}

	/* initialize module */
	DBG_MSG(3, "Initializing module...");
	if (module_callback(mod, mod_id, CB_INITIALIZE, mod->info[mod_id].var_offset, mod->info[mod_id].args) != 0) {
		fprintf(stderr, "Module \"%s\" initialization failed.\n", mod->info[mod_id].name);
		module_remove(mod, mod_id);
		return -1;
	}
	DBG_MSG(3, "Initializing module... done.");


	DBG_MSG(2, "Module loaded (mod_id = %d, real name = \"%s\").", mod_id, mod->info[mod_id].real_name);
	return mod_id;
}


/*****************************************************************************
 * module_remove
 */
int module_remove(MODULE *mod, int mod_id) {
	assert(mod != NULL);
	assert(mod_id < mod->info_count);

	if (mod->info[mod_id].name) {
		DBG_MSG(1, "Removing module (id = %d)", mod_id);

		if (mod->info[mod_id].dlh != NULL)
			module_callback(mod, mod_id, CB_CLEANUP, 0);

		mod->info[mod_id].data = NULL;
		mod->info[mod_id].real_name = NULL;

		free(mod->info[mod_id].name);

		if (mod->info[mod_id].args)
			free(mod->info[mod_id].args);

		if (mod->info[mod_id].cfg_fd_old)
			free(mod->info[mod_id].cfg_fd_old);

		if (mod->info[mod_id].dlh) {
			if (dlclose(mod->info[mod_id].dlh)) {
				fprintf(stderr, "Can't unload module \"%s\": %s",
						mod->info[mod_id].name, dlerror());
			}
		}

		DBG_MSG(2, "Module removed.");
	}
	return 0;
}


/*****************************************************************************
 * module_load_all
 */
int  module_load_all(MODULE *mod) {
	int id = 0;

	assert(mod != NULL);

	DBG_MSG(1, "Loading all modules ...");

	for (id = 0; id < mod->info_count; id++) {
		if (mod->info[id].name && !mod->info[id].real_name) {
			if (module_load(mod, id) == -1) {
				return -1;
			}
		}
	}
	return 0;

	DBG_MSG(2, "All modules loaded.");
}


/*****************************************************************************
 * module_remove_all
 */
void module_remove_all(MODULE *mod) {
	int id = 0;
	assert(mod != NULL);
	DBG_MSG(1, "Removing all modules ...");
	for (id = 0; id < mod->info_count; id++)
		module_remove(mod, id);
	DBG_MSG(2, "All modules removed.");
}


/*****************************************************************************
 * module_callback
 */
int module_callback(MODULE *mod, int mod_id, int callback, ...) {
	int ret = 0;
	va_list ap;
	char *str;
	int n, ***ppp_i, **p_n;
	long long l, **pp_ll;

	assert(mod != NULL);
	assert(mod_id >= 0);
	assert(mod_id < mod->info_count);
	assert(callback >= 0);
	assert(callback < CB_MAX);
	
	
	/* check if module is loaded */
	if (!mod->info[mod_id].name) {
		DBG_MSG(1, "Module %d not loaded.", mod_id);
		return -1;
	}

	DBG_MSG(2, "Callback \"%s\" for module \"%s\"(%d) ...",
			cb_name[callback], mod->info[mod_id].name, mod_id);

	/* load callback symbol */
	if (mod->info[mod_id].callback[callback] == NULL) {
		mod->info[mod_id].callback[callback] = module_lookup(mod, mod_id, cb_name[callback]);
	}

	/* call callback */
	if (mod->info[mod_id].callback[callback]) {
		DBG_MSG(2, "Launching ... ");
		va_start(ap, callback);
		switch (callback) {
			case CB_INITIALIZE:
				n = va_arg(ap, int);
				str = va_arg(ap, char*);
				DBG_MSG(3, "Callback: %s, %d, \"%s\"", cb_name[callback], n, str);
				ret = (*((int(*)(int, const void **, int, const char*))mod->info[mod_id].callback[callback]))
						(mod_id, (const void**) &(mod->info[mod_id].data), n, (const char*)str);
				break;

			case CB_CLEANUP:
			case CB_TIMEOUT:
				ret = (*((int(*)(int, const void **))mod->info[mod_id].callback[callback]))
						(mod_id, (const void**) &(mod->info[mod_id].data));
				break;

			case CB_FILEDESCRIPTOR:
				n = va_arg(ap, int);
				ret = (*((int(*)(int, const void **, int))mod->info[mod_id].callback[callback]))
						(mod_id, (const void**) &(mod->info[mod_id].data), n);
				break;
			case CB_GET_TIMEOUT_PTR:
				pp_ll = va_arg(ap, long long **);
				p_n = va_arg(ap, int **);
				ret = (*((int(*)(int, const void **, long long **, int **))
						mod->info[mod_id].callback[callback]))
						(mod_id, (const void**) &(mod->info[mod_id].data),
						(long long **)pp_ll, (int **)p_n);
				break;
			case CB_GET_FD_LIST_PTR:
				ppp_i = va_arg(ap, int ***);
				l = va_arg(ap, long long);
				ret = (*((int(*)(int, const void **, int ***))mod->info[mod_id].callback[callback]))
						(mod_id, (const void**) &(mod->info[mod_id].data), (int ***)ppp_i);
				break;
		}
		va_end(ap);
		DBG_MSG(2, "Launching ... returned %d", ret);
	} else {
		ret = -1;
	}

	DBG_MSG(2, "Callback \"%s\" for module \"%s\"(%d) ... done (returned %d)",
			cb_name[callback], mod->info[mod_id].name, mod_id, ret);

	return ret;
}


/*****************************************************************************
 * module_fill_fdset
 */
int module_fill_fdset(MODULE *mod, fd_set *s) {
	int id;
	long size;
	int *p;

	assert(mod != NULL);

	for (id = 0; id < mod->info_count; id++) {
		if (mod->info[id].name) {
			if (!mod->info[id].cfg_fd) {
				//mod->info[id].cfg_fd = module_lookup(mod, id, "mod_filedescriptor_value");
				module_callback(mod, id, CB_GET_FD_LIST_PTR, &(mod->info[id].cfg_fd));
			}
			if (mod->info[id].cfg_fd && *(int**)(mod->info[id].cfg_fd)) {
				size = 1;
				for (p = *(int **) mod->info[id].cfg_fd; p && *p != -1; p++) {
					FD_SET(*p, s);
					DBG_MSG(8, "Registered fd %d from module \"%s\" at pos. %ld",
							*p, mod->info[id].name, size - 1);
					size++;
				}
				DBG_MSG(7, "Registered %ld fds.", size - 1);
				if (mod->info[id].cfg_fd_old)
					free(mod->info[id].cfg_fd_old);

				mod->info[id].cfg_fd_old = (int *) malloc(sizeof(int) * size);
				memcpy(mod->info[id].cfg_fd_old, *(int***)(mod->info[id].cfg_fd), (sizeof(int) - 1) * size);
				mod->info[id].cfg_fd_old[size - 1] = -1;
			}
		}
	}
	return 0;
}


/*****************************************************************************
 * module_handle_fds
 */
void module_handle_fds(MODULE *mod, fd_set *s) {
	int id, i, fd;
	
	assert(mod != NULL);

	for (id = 0; id < mod->info_count; id++) {
		if (mod->info[id].name && mod->info[id].cfg_fd_old) {
			for (i = 0; (fd = mod->info[id].cfg_fd_old[i]) != -1; i++) {
				if (FD_ISSET(fd, s)) {
					DBG_MSG(1, "Event on fd %d", fd);
					module_callback(mod, id, CB_FILEDESCRIPTOR, fd);
				}
			}

			free(mod->info[id].cfg_fd_old);
			mod->info[id].cfg_fd_old = NULL;
		}
	}
}


/*****************************************************************************
 * module_handle_timeout
 */
long long module_handle_timeout(MODULE *mod, long long timeout_duration, int important_only, int force_nonimportant) {
	int id, force;
	long long *t = NULL;
	long long t_min = -1;

	assert(mod != NULL);
	assert(timeout_duration >= 0);

	DBG_MSG(3, "Timeout %1.1f ms ...", timeout_duration/1e3f);
	for (id = 0; id < mod->info_count; id++) {
		if (mod->info[id].name) {
			if (!mod->info[id].cfg_timeout) {
				module_callback(mod, id, CB_GET_TIMEOUT_PTR,
						&(mod->info[id].cfg_timeout), &(mod->info[id].cfg_timeout_always));
			}
			if ((mod->info[id].cfg_timeout && mod->info[id].cfg_timeout_always)
					&& ((*(mod->info[id].cfg_timeout_always)) || !important_only)) {
				
				t = *(long long**)(&(mod->info[id].cfg_timeout));
				force = (force_nonimportant && !(*(mod->info[id].cfg_timeout_always)));

				if (*t > 0 || force) {
					*t = force ? 0 : *t - timeout_duration;
					DBG_MSG(4, "Module \"%s\": Timeout after %1.1f ms", mod->info[id].name, (*t)/1e3f);
					if (*t <= 0) {
						/* timed out */
						*t = 0;
						module_callback(mod, id, CB_TIMEOUT, 0);
					}
					if (*t > 0 && (*t < t_min || t_min == -1))
						t_min = *t;
				}
			}
		}
	}

	DBG_MSG(4, "Next timeout after %1.1f ms", t_min/1e3f);
	return t_min == -1 ? 0 : t_min;
}

