/*
 * battery.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "battery.h"

#define DEBUG 0
#include "debug.h"

#define mod_name	"battery.c"
#define MAX_LINE_LEN	80

#define FILENAME_STATE	"/proc/acpi/battery/%s/state"
#define FILENAME_INFO	"/proc/acpi/battery/%s/info"
#define FILENAME_LEN	60

int read_battery(const char *battery, int *r_seconds, float *r_part, int *r_state) {
	FILE *f = NULL;
	int val, capacity, rem_capacity, rate;
	char line[MAX_LINE_LEN], key[MAX_LINE_LEN], str_val[MAX_LINE_LEN];
	char *fn;
	int have_rem_capacity = 0, have_state = 0, have_rate = 0;

	assert(battery != NULL);
	assert(r_seconds != NULL);
	assert(r_part != NULL);
	assert(r_state != NULL);
	
	DBG_MSG(1, "Reading battery state ...");

	fn = (char *) malloc(FILENAME_LEN);
	snprintf(fn, FILENAME_LEN, FILENAME_STATE, battery);

	*r_state = BAT_MISSING;
	if ((f = fopen(fn, "r"))) {
		while (!(have_rate && have_state && have_rem_capacity) && fgets(line, MAX_LINE_LEN, f)) {
			sscanf(line, "%[^:]: %s", key, str_val);
			DBG_MSG(2, "%s: \"%s\" = \"%s\"", fn, key, str_val);

			if (!have_state && strcmp(key, "charging state") == 0) {
				*r_state = BAT_UNKNOWN;
				if (strcmp(str_val, "discharging") == 0)
					*r_state = BAT_DISCHARGING;
				if (strcmp(str_val, "charged") == 0)
					*r_state = BAT_CHARGED;
				if (strcmp(str_val, "charging") == 0)
					*r_state = BAT_CHARGING;
				have_state = 1;
			} else {
				if (isdigit(str_val[0])) {
					val = atoi(str_val);
				} else {
					val = 0;
				}
				DBG_MSG(2, "%s: \"%s\" = \"%s\" (= %d)", fn, key, str_val, val);

				if (!have_rem_capacity && strcmp(key, "remaining capacity") == 0) {
					rem_capacity = val;
					DBG_MSG(2, "capacity += %d", val);
					have_rem_capacity = 1;
				} else if (!have_rate && strcmp(key, "present rate") == 0) {
					rate = val;
					DBG_MSG(2, "rate += %d", val);
					have_rate = 1;
				}
			}
		}
		fclose(f);

		if (*r_state == BAT_CHARGING || *r_state == BAT_DISCHARGING) {
			capacity = 0;
			/* read battery total capacity */
			snprintf(fn, FILENAME_LEN, FILENAME_INFO, battery);
			if ((f = fopen(fn, "r"))) {
				while (fgets(line, MAX_LINE_LEN, f)) {
					sscanf(line, "%[^:]: %s", key, str_val);
					DBG_MSG(2, "%s: \"%s\" = \"%s\"", fn, key, str_val);
					if (strcmp(key, "last full capacity") == 0) {
						capacity = atoi(str_val);
						break;
					}
				}
				fclose(f);
			}
		}
	} else {
		free(fn);
		DBG_MSG(1, "Reading battery state ... Failed.");
		return 0;
	}

	if (rem_capacity == 0) {
		*r_state = BAT_MISSING;
		*r_seconds = -1;
		*r_part = 0.0;
	} else {
		if (*r_state == BAT_CHARGED) {
			*r_part = 1.0;
			*r_seconds = -1;
		} else {
			if (*r_state == BAT_DISCHARGING) {
				*r_seconds = (int)((float)rem_capacity / (float)rate * 3600);
			} else {
				*r_seconds = (int)((float)(capacity - rem_capacity) / (float)rate * 3600);
			}
			*r_part = (float) rem_capacity / (float)capacity;
		}
	}
	free(fn);
	DBG_MSG(1, "Reading battery state ... Done: seconds = %d, part = %0.1f, state = %d",
			*r_seconds, *r_part, *r_state);
	return 1;
}

