/*
 * main.c - icebuteo module "Clock"
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

/* include icebuteo module API */
#include <icebuteo-api.h>

#include "wmcdclock.h"

/* debug */
#define DEBUG 2
#include "debug.h"

/*
 * Importat details:
 *
 *	- There is no main.
 *	- Everything is callback.
 *	- Each callback gets module id (mod_id), pointer to module's data (mod_data)
 *	  and one integer parameter.
 *	- All modules share one thread, so callback must exit as soon as
 *	  possible.
 *	- All callbacks have same prototypes.
 *	- stdin, stdout and stderr is unchanged.
 *	- Module can be loaded multiple times, so do NOT use global variables,
 *	  use mod_data structure.
 *
 */

/* module real name */
const char *mod_name = "Clock";

#define FIFO_BUF_MAX	128

/* module data */
typedef struct {
	/* important configuration */
	int var_offset;			/* first variable offset */
	long long mod_timeout_value;	/* timeout settings (0 = never) */
	int mod_timeout_always;		/* timeout is required even if window is hidden */
	int *mod_filedescriptor_value;	/* array of filedescriptors (terminated by -1) */

	/* other variables put here */
	time_t time_old;
	time_t count_down_time;
	time_t count_down_start;
	int count_down;
	int zero_blink;
	int fifo_buf_size;
	char fifo_buf[FIFO_BUF_MAX];

	/* configuration */
	char *date_format;
	char *time_format;
	char *wmcdclock_format;
	char *wmcdclock_fifo;
	int update_seconds;
	int date_var;
	int time_var;
	int wmcdclock_var_float;
	int wmcdclock_var_string;
} MOD_DATA;

/* macro for easy access to module's data */
#define MD	((MOD_DATA *) (*mod_data))


/*****************************************************************************
 * kernel-module data exchange functions
 *
 * All of this ones returns pointer in parameter. Return value is unused.
 * Just only return pointer, don't do anything else.
 */
int mod_get_timeout_ptr(int mod_id, const void **mod_data, long long **value, int **timeout_always) {
	*value = &(MD->mod_timeout_value);
	*timeout_always = &(MD->mod_timeout_always);
	return 0;
}
int mod_get_fd_list_ptr(int mod_id, const void **mod_data, int ***list_begin) {
	*list_begin = &(MD->mod_filedescriptor_value);
	return 0;
}

/*****************************************************************************
 * Initialize callback
 *
 * Called once after module is loaded.
 * Returns zero if success.
 */
int mod_initialize(int mod_id, const void **mod_data, int var_offset, const char *arg) {
	const int cfg_tr_len = 9;
	int i = 1;
	P_TRANS cfg_tr[] = {
		/* const. name,			id, var, type */
		{ "date-format",		1, NULL, P_STRING },
		{ "date-var",			2, NULL, P_INTEGER },
		{ "time-format",		3, NULL, P_STRING },
		{ "time-var",			4, NULL, P_INTEGER },
		{ "update-seconds",		9, NULL, P_INTEGER },
		{ "wmcdclock-fifo",		5, NULL, P_STRING },
		{ "wmcdclock-format",		6, NULL, P_STRING },
		{ "wmcdclock-var-float",	7, NULL, P_INTEGER },
		{ "wmcdclock-var-string",	8, NULL, P_INTEGER },
		{ NULL, 0, NULL, 0 }
	};
	static P_FORMAT cfg_format[] = {
		P_CONSTANT, P_EQUAL, P_STRING | P_INTEGER | P_FLOAT, 0
	};

	DBG_MSG(1, "initialize...");

	/* alloc memory for "global" variables */
	*mod_data = malloc(sizeof(MOD_DATA));
	memset((void *)*mod_data, 0, sizeof(MOD_DATA));

	/* save offset */
	MD->var_offset = var_offset;

	/* load configuration */
	i = 0;
	cfg_tr[i++].var = &(MD->date_format);
	cfg_tr[i++].var = &(MD->date_var);
	cfg_tr[i++].var = &(MD->time_format);
	cfg_tr[i++].var = &(MD->time_var);
	cfg_tr[i++].var = &(MD->update_seconds);
	cfg_tr[i++].var = &(MD->wmcdclock_fifo);
	cfg_tr[i++].var = &(MD->wmcdclock_format);
	cfg_tr[i++].var = &(MD->wmcdclock_var_float);
	cfg_tr[i++].var = &(MD->wmcdclock_var_string);
	if (!parse_file(arg, NULL, cfg_format, -1, 0, 2, cfg_tr, cfg_tr_len)) {
		return -1;
	}

	/* register timeout - first after 0.1 seconds */
	MD->mod_timeout_value = 100e3;
	MD->mod_timeout_always = 0;

	/* the init */
	MD->time_old = 1;
	MD->count_down = 0;
	MD->count_down_time = 0;

	/* init vars */
	var_set_as_int(MD->var_offset + 0, 1);
	var_set_as_string(MD->var_offset + MD->date_var, "");
	var_set_as_string(MD->var_offset + MD->time_var, "");
	var_set_as_string(MD->var_offset + MD->wmcdclock_var_string, "");
	var_set_as_float(MD->var_offset + MD->wmcdclock_var_float, 0.0);

	/* open fifo */
	MD->mod_filedescriptor_value = (int *)malloc(sizeof(int) * 2);
	MD->mod_filedescriptor_value[1] = -1;
	if (!(MD->mod_filedescriptor_value[0] = open_fifo(MD->wmcdclock_fifo)))
		return -1;

	return 0;
}


/*****************************************************************************
 * Cleanup callback
 *
 * Called once before module is unloaded.
 * Parameter zero is unused (always zero).
 */
int mod_cleanup(int mod_id, const void **mod_data) {
	DBG_MSG(1, "cleanup...");
	close_fifo(MD->wmcdclock_fifo, MD->mod_filedescriptor_value[0]);
	free(MD);
	if(MD->mod_filedescriptor_value)
		free(MD->mod_filedescriptor_value);
	return 0;
}


/*****************************************************************************
 * Timeout callback
 */
int mod_timeout(int mod_id, const void **mod_data) {
	time_t time_new = 1, time_show, secs_rem;
	char outstr[256];
	struct tm tmp, cd;
	float part;

	MD->mod_timeout_value = 333e3;

	DBG_MSG(1, "Updating time ...");

	time_new = time(NULL);
	if (!(MD->count_down || MD->update_seconds)) {
		secs_rem = time_new % 60;
		time_new -= secs_rem;

		if (secs_rem <= 58) {
			MD->mod_timeout_value = (59 - secs_rem) * 1e6;
			DBG_MSG(2, "Next update after %0.2f s.", MD->mod_timeout_value / 1.0e6);
		}
	}

	if (time_new != MD->time_old) {
		/* update vars */
		DBG_MSG(3, "Tik");
		tmp = *localtime(&time_new);
		strftime(outstr, sizeof(outstr), MD->time_format, &tmp);
		var_set_as_string(MD->var_offset + MD->time_var, outstr);
		strftime(outstr, sizeof(outstr), MD->date_format, &tmp);
		var_set_as_string(MD->var_offset + MD->date_var, outstr);
		MD->time_old = time_new;

		/* update countdown */
		if (MD->count_down) {
			DBG_MSG(3, "update count down");
			time_show = MD->count_down_time - time_new;
			cd = *gmtime(&time_show);
			
			strftime(outstr, sizeof(outstr), MD->wmcdclock_format, &cd);
			var_set_as_string(MD->var_offset + MD->wmcdclock_var_string, outstr);

			part = 1.0 - ((float)time_show / (float)(MD->count_down_time - MD->count_down_start));
			DBG_MSG(3, "part = %0.3f, time_show = %ld, len = %ld",
					part, time_show, MD->count_down_time - MD->count_down_start);
			var_set_as_float(MD->var_offset + MD->wmcdclock_var_float, part);

			if (MD->count_down_time <= time_new) {
				MD->count_down = 0;
				MD->zero_blink = 30;
			}
		} else {
			var_set_as_string(MD->var_offset + MD->wmcdclock_var_string, "");
		}
	}
	if (MD->count_down == 0 && MD->zero_blink > 0) {
		MD->zero_blink--;
		if (MD->zero_blink % 2) {
			time_show = 0;
			cd = *gmtime(&time_show);
			strftime(outstr, sizeof(outstr), MD->wmcdclock_format, &cd);
			var_set_as_string(MD->var_offset + MD->wmcdclock_var_string, outstr);
			var_set_as_float(MD->var_offset + MD->wmcdclock_var_float, 1.0);
		} else {
			var_set_as_string(MD->var_offset + MD->wmcdclock_var_string, "");
			var_set_as_float(MD->var_offset + MD->wmcdclock_var_float, 0.0);
		}
		MD->mod_timeout_value = 100e3;
	}
	return 0;
}


/*****************************************************************************
 * filedescriptor callback
 */
int mod_filedescriptor(int mod_id, const void **mod_data, int fd) {
	int rd;
	char buf[31];

	DBG_MSG(2, "file descriptor callback...");
	if ((rd = read(MD->mod_filedescriptor_value[0], buf, 31)) == 0 || rd + MD->fifo_buf_size > FIFO_BUF_MAX) {
		DBG_MSG(2, "wmcdclock command: \"%s\"", MD->fifo_buf);

		/* prepare fifo for next client*/
		MD->mod_filedescriptor_value[0] = reopen_fifo(MD->wmcdclock_fifo, MD->mod_filedescriptor_value[0]);

		/* start countdown */
		MD->count_down = parse_fifo(MD->fifo_buf, &(MD->count_down_time));
		MD->count_down_start = time(NULL);
		MD->mod_timeout_value = 333e3;

		/* prepare buffer */
		MD->fifo_buf_size = 0;
		MD->fifo_buf[0] = 0;
	} else {
		buf[rd] = '\0';
		strcat(MD->fifo_buf, buf);
		MD->fifo_buf_size += rd;
	}
	DBG_MSG(2, "file descriptor callback... Done.");
	return 0;
}

