/*
 * main.c - icebuteo module "Skeleton"
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

/* include icebuteo module API */
#include <icebuteo-api.h>

/* debug */
#define DEBUG 5
#include "debug.h"

/*
 * Importat details:
 *
 *	- There is no main.
 *	- Everything is callback.
 *	- Each callback gets module id (mod_id), pointer to module's data (mod_data)
 *	  and sometimes something else :)
 *	- All modules share one thread, so callback must exit as soon as
 *	  possible.
 *	- All callbacks have same prototypes.
 *	- stdin, stdout and stderr is unchanged.
 *	- Module can be loaded multiple times, so do NOT use global variables,
 *	  use mod_data structure.
 *
 */

/* module name */
const char *mod_name = "Skeleton";

/* module data */
typedef struct {
	/* important configuration */
	int var_offset;			/* first variable offset */
	long long mod_timeout_value;	/* timeout settings (0 = never) */
	int mod_timeout_always;		/* timeout is required even if window is hidden */
	int *mod_filedescriptor_value;	/* array of filedescriptors (terminated by -1) */

	/* other variables put here */
	int the_time;

} MOD_DATA;

/* macro for easy access to module's data */
#define MD	((MOD_DATA *) (*mod_data))

/*****************************************************************************
 * kernel-module data exchange functions
 *
 * All of this ones returns pointer in parameter. Return value is unused.
 * Just only return pointer, don't do anything else.
 */
int mod_get_timeout_ptr(int mod_id, const void **mod_data, long long **value, int **timeout_always) {
	*value = &(MD->mod_timeout_value);
	*timeout_always = &(MD->mod_timeout_always);
	return 0;
}
int mod_get_fd_list_ptr(int mod_id, const void **mod_data, int ***list_begin) {
	*list_begin = &(MD->mod_filedescriptor_value);
	return 0;
}


/*****************************************************************************
 * Initialize callback
 *
 * Called once after module is loaded.
 * Returns zero if success.
 */
int mod_initialize(int mod_id, const void **mod_data, int var_offset, const char *arg) {
	int n;

	DBG_MSG(1, "initialize ...");
	DBG_MSG(2, "var_offset = %d", var_offset);
	DBG_MSG(2, "arg = \"%s\"", arg);

	/* alloc memory for "global" variables */
	*mod_data = malloc(sizeof(MOD_DATA));

	/* save offset */
	MD->var_offset = var_offset;

	/* init other variables */
	MD->the_time = 0;

	/* examples: */
	var_set_as_string(MD->var_offset + 1, (char *)mod_name);
	n = var_get_as_int(MD->var_offset);
	DBG_MSG(1, "var[%d] = %d", MD->var_offset, n);

	/* register stdin */
	MD->mod_filedescriptor_value = (int *) malloc(sizeof(int) * 2);
	MD->mod_filedescriptor_value[0] = 0;
	MD->mod_filedescriptor_value[1] = -1;

	/* register timeout - first after 1 seconds */
	MD->mod_timeout_value = 1000e3;

	return 0;
}


/*****************************************************************************
 * Cleanup callback
 *
 * Called once before module is unloaded.
 * Parameter zero is unused (always zero).
 */
int mod_cleanup(int mod_id, const void **mod_data) {
	DBG_MSG(1, "cleanup...");
	free(MD);
	if (MD->mod_filedescriptor_value)
		free(MD->mod_filedescriptor_value);
	return 0;
}


/*****************************************************************************
 * Timeout callback
 */
int mod_timeout(int mod_id, const void **mod_data) {
	DBG_MSG(1, "timeout - next in 1 second");
	MD->mod_timeout_value = 1e6;
	MD->the_time++;
	var_set_as_int(MD->var_offset + 1, MD->the_time);
	return 0;
}


/*****************************************************************************
 * filedescriptor callback
 */
int mod_filedescriptor(int mod_id, const void **mod_data, int fd) {
	char buf[256], *p;
	DBG_MSG(1, "filedescriptor");
	if (read(fd, buf, 256) > 0) {
		p = strchr(buf, '\n');
		if (p) {
			*p = 0;
			var_set_as_string(MD->var_offset + 2, buf);
			DBG_MSG(2, "received \"%s\"", buf);
		}
	} else {
		DBG_MSG(1, "eof, closing fd...");
		close(fd);
		free(MD->mod_filedescriptor_value);
		MD->mod_filedescriptor_value = NULL;
	}
	return 0;
}


/*****************************************************************************
 * Self-test main.
 */
#ifdef SELFTEST
int main (void) {
	MOD_DATA *data = NULL;
	int mod_id = 1;

	mod_initialize(mod_id, &data);
	mod_cleanup(mod_id, &data);

	return EXIT_SUCCESS;
}
#endif

