/*
 * parser.c
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <locale.h>

#include "parser.h"
#include "parser-lex.h"

#define DEBUG 0
#include "debug.h"

P_LINE *parse_line(const char *filename, const FILE *f, P_FORMAT *format, int format_len) {
	P_LINE *ln;
	int t, i;
	int pos;
	char *p;

	assert(filename != NULL);
	assert(f != NULL);
	assert(format != NULL);
	assert(format_len > 0 || format_len == -1);

	/* calculate format length if not specified */
	if (format_len == -1) {
		for (format_len = 0; format[format_len] != 0; format_len++)
			;
	}

	/* check format_len */
	assert(format[format_len] == 0);

	/* alloc. line */
	ln = (P_LINE*) malloc(sizeof(P_LINE));
	ln->token = (P_TOKEN *) malloc(sizeof(P_TOKEN) * format_len);
	ln->size = format_len;
	ln->line = 0;
	for (i = 0; i < ln->size; i++) {
		ln->token[i].val = NULL;
	}

	/* parse line & store into ln */
	parser_lex_in = (FILE *)f;
	pos = 0;
	while ((t = parser_lex_scan(filename)) > 0) {
		if (t == P_EOL) {
			if (pos == 0) {
				/* ignore emty / commented lines */
				continue;
			} if (pos == ln->size) {
				/* line complete */
				DBG_MSG(1, "Line parsed.");
				return ln;
			} else {
				fprintf(stderr, "%s: %d: Unexpected end of line.\n", filename, parser_lex_lineno);
				break;
			}
		}
		
		DBG_MSG(3, "line %d: value: %s, type %d", parser_lex_lineno, parser_lex_text, t);

		/* compare readed data with specified format */
		if ((format[pos] & t) != 0 && pos < ln->size) {
			ln->token[pos].type = t;
			switch (t) {
				case P_NOTHING:
				case P_EQUAL:
				case P_CONSTANT:
					ln->token[pos].val_int = 0;
					ln->token[pos].val_float = 0.0;
					ln->token[pos].val = strdup(parser_lex_text);
					break;

				case P_STRING:
					ln->token[pos].val_int = 0;
					ln->token[pos].val_float = 0.0;
					ln->token[pos].val = strdup(parser_lex_text + 1);
					ln->token[pos].val[strlen(ln->token[pos].val) - 1] = 0;
					p = ln->token[pos].val;
					while ((p = strchr(p, '\\'))) {
						if (*(p + 1) == '"') {
							memmove(p, p + 1, strlen(p));
						}
					}
					break;

				case P_FLOAT:
					ln->token[pos].val_int = (int) atof(parser_lex_text);
					ln->token[pos].val_float = atof(parser_lex_text);
					ln->token[pos].val = strdup(parser_lex_text);
					break;

				case P_INTEGER:
					ln->token[pos].val_int = atoi(parser_lex_text);
					ln->token[pos].val_float = atof(parser_lex_text);
					ln->token[pos].val = strdup(parser_lex_text);
					break;
			}
		} else {
			fprintf(stderr, "%s: %d: Unexpected \"%s\"\n", filename, parser_lex_lineno, parser_lex_text);
			break;
		}
		pos++;
		
	}
	if (t == 0)
		DBG_MSG(1, "End-of-file");

	parse_free_line(ln);
	return NULL;
}

void parse_free_line(P_LINE *ln) {
	int i;
	for (i = 0; i < ln->size; i++) {
		if (ln->token[i].val)
			free(ln->token[i].val);
	}
	free(ln);
}

int parse_file(const char *filename, const FILE *f, P_FORMAT *format, int format_len,
		int symbol_pos, int value_pos, const P_TRANS *def, size_t def_len) {
	P_LINE *ln = NULL;
	FILE *F;
	int ret = 1;

	assert(filename != NULL);
	assert(format != NULL);
	assert(format_len > 0 || format_len == -1);
	assert(def != NULL);
	assert(def_len > 0);

	DBG_MSG(1, "Parsing file \"%s\" ...", filename);

	if (f == NULL) {
		F = fopen(filename, "r");
		if (!F) {
			perror(filename);
			return 0;
		}
	} else {
		F = (FILE *) f;
	}

	while ((ln = parse_line(filename, F, format, format_len))) {
		DBG_MSG(2, "Option %s = %s", ln->token[symbol_pos].val, ln->token[value_pos].val);
		if (parse_set_var_by_name(def, def_len, ln->token[symbol_pos].val,
				(ln->token[2].type == P_INTEGER)
					? (void *) &(ln->token[value_pos].val_int)
					: ((ln->token[value_pos].type == P_FLOAT)
						? (void *) &(ln->token[value_pos].val_float)
						: (void *) &(ln->token[value_pos].val)),
				ln->token[value_pos].type, 0) == -1)
		{
			DBG_MSG(1, "Parsing file \"%s\" ... Failed.", filename);
			ret = 0;
			break;
		}
		parse_free_line(ln);
	}
	if (f == NULL) {
		if (fclose(F) != 0)
			perror(filename);
	}
	if (ret == 1)
		DBG_MSG(1, "Parsing file \"%s\" ... Done.", filename);
	return ret;
}

/****************************************************************************/

static int parse_trans_cmp(const void *a, const void *b) {
	P_TRANS *pa = (P_TRANS*)a;
	P_TRANS *pb = (P_TRANS*)b;
	return strcmp(pa->str, pb->str);
}

int parse_set_var_by_name(const P_TRANS *def, size_t def_len, const char *name, void *value, P_FORMAT value_type, int use_free) {
	P_TRANS key;
	P_TRANS *result = NULL;
	assert(def != NULL);
	assert(def_len > 0);
	assert(name != NULL);
	assert(value_type == (value_type & (P_INTEGER | P_STRING | P_FLOAT)));

	key.str = (char *)name;
	result = (P_TRANS *) bsearch(&key, def, def_len, sizeof(P_TRANS), parse_trans_cmp);
	if (value && value_type && result) {

		if (result->type != value_type) {
			fprintf(stderr, "Error: Type mismatch - %s\n", name);
			return -1;
		}

		switch(value_type) {
			case P_INTEGER:
				*((int*)(result->var)) = *((int*)value);
				break;
			case P_FLOAT:
				*(((float*)result->var)) = *((float*)value);
				break;
			case P_STRING:
				DBG_MSG(2, "var: %s", *((char**)result->var));
				if(use_free && *((char**)result->var)) {
					DBG_MSG(3, "freeing old string...");
					free(*((char**)result->var));
				}
				*((char**)result->var) = strdup(*((char**)value));
				break;
		}
	}
	if (!result)
		fprintf(stderr, "Error: Symbol \"%s\" not found\n", name);

	return result ? result->id : -1;
}

